import logging

import torch
import numpy as np

from ._core import LukasiewiczCore


class BasePredicates(LukasiewiczCore):

    """
    Class for Predicates.  Aid in explanation.
    """

    def __init__(self, feature_names):
        super(BasePredicates, self).__init__()
        self.operands = feature_names
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _produce_negation_string(
            weight: float,
            text: str,
            explain_type: str = 'both',
            print_type: str = 'logical',
            negate: bool = False,
            **kwargs
    ) -> str:
        """
        Produce a negation under certain conditions.

        Args:
            weight (float): Weight for specific predicate.
            text(str): Text for specific predicate.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.

        Returns:
            str: The text with negation applied.
        """
        assert print_type in ['logical', 'logical-natural', 'natural'], \
            "'print_type' must be one of 'logical', 'logical-natural', or 'natural'"

        # conditions that produce a negative
        if (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'negative'):
            if print_type == 'logical':
                return f"NOT({text})".replace("NOT(\n", "\nNOT(")
            elif print_type == 'logical-natural' or print_type == 'natural':
                text = f"NOT {text}"
                return f"{text}"
            else:
                raise ValueError("`print_type` must be one of `logical`, `logical-natural`, or `natural`")
        elif (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and explain_type == 'positive':
            return ""
        # conditions that produce a positive
        if (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'positive'):
            return text
        elif (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and explain_type == 'negative':
            return ""
        else:
            return text

    def _produce_predicate_value(
            self,
            name: str,
            weight: float,
            input_value: float,
            required_output_threshold: float,
            ignore_uninformative: bool,
            print_type: str = 'logical',
            rounding_precision: int = 3,
            negate: bool = False,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> str:
        """
        Produce the string for predicate explanation.

        Args:
            name (str): predicate name.
            weight (float): predicates weight.
            required_output_threshold (float): required output below which predicate will be excluded from explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic

        Returns:
            str: predicate with value explanation.
        """
        # required_output_threshold represents the threshold for each predicate with all negations applied.
        # So we have the following scenarios:
            # if negate == True:
                # input feature negated <= required_output_threshold
                #
                # if w < 0:
                    # 1 - x <= required_output_threshold
                    # x >= 1 - required_output_threshold
                # else:
                    # x <= required_output_threshold
                    # NOT(x >= required_output_threshold)
            # else:
                # input feature negated >= threshold
                #
                # if w < 0:
                    # 1 - x >= required_output_threshold
                    # x <= 1 - required_output_threshold
                    # NOT(x >= 1 - required_output_threshold)
                # else:
                    # x >= required_output_threshold

        if not negate:
            if ((np.isclose(round(required_output_threshold, rounding_precision), 0.0) and weight >= 0)
                or (np.isclose(round(1.0 - required_output_threshold,
                          rounding_precision), 1.0) and weight <= 0)) and ignore_uninformative:
                return ""
        else:
            if ((np.isclose(round(1.0 - required_output_threshold, rounding_precision), 0.0) and weight <= 0)
                or (np.isclose(round(required_output_threshold,
                          rounding_precision), 1.0) and weight >= 0)) and ignore_uninformative:
                return ""

        if print_type == 'logical':
            gte_str = ">="
        elif print_type == 'logical-natural' or print_type == 'natural':
            gte_str = "greater than or equal to"
        else:
            raise AssertionError("print_type must be 'logical', 'logical-natural', or 'natural'.")

        # if a search has been performed still display using the originally requested rounding precision
        rounding_precision = kwargs.get('original_rounding_precision', rounding_precision)
        if inverse_transform is not None:
            input_array = np.zeros((1, len(self.operands)))
            input_array_actuals = np.zeros((1, len(self.operands)))
            idx = self.operands.index(name)
            if weight >= 0:
                input_array[0, idx] = required_output_threshold
            else:
                input_array[0, idx] = 1.0 - required_output_threshold
            transformed_output_threshold = inverse_transform(input_array)[0, idx]

            input_array_actuals[0, idx] = input_value
            transformed_input_value = inverse_transform(input_array_actuals)[0, idx]

            if show_bounds:
                # return f"{name} {gte_str} {round(transformed_output_threshold, rounding_precision)} [actual: {round(transformed_input_value, rounding_precision)}]"
                return f"{name} {gte_str} {round(transformed_output_threshold, rounding_precision)}"
            return f"{name}"

        if weight >= 0:
            if show_bounds:
                # if negate is true then produce_negation_string will wrap
                # this in NOT() which is equivalent to the LTE condition
                # TODO: do we want to show the actuals?
                # return f"{name} {gte_str} {round(required_output_threshold, rounding_precision)} [actual: {round(input_value, rounding_precision)}]"
                return f"{name} {gte_str} {round(required_output_threshold, rounding_precision)}"
            return f"{name}"
        if show_bounds:
            # if negate is true then produce_negation_string will
            # wrap this in NOT() which is equivalent to the LTE condition
            # TODO: do we want to show the actuals?
            # return f"{name} {gte_str} {round(1 - required_output_threshold, rounding_precision)} [actual: {round(input_value, rounding_precision)}]"
            return f"{name} {gte_str} {round(1 - required_output_threshold, rounding_precision)}"
        return f"{name}"

    def explain_sample(
            self,
            quantile: float = 0.5,
            required_output_thresholds: torch.Tensor = None,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            explain_type: str = 'both',
            parent_logic_type: str = None,
            negate: bool = False,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            depth: int = 0,
            print_explanation: bool = False,
            global_explain: bool = False,
            ignore_uninformative: bool = False,
            print_type: str = 'logical',
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            feature_importances: bool = False,
            feature_importances_type: str = '',
            **kwargs
    ) -> list:
        """
        Produce a sample explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            global_explain (bool): If True, perform a global explanation.
            print_explanation (bool): If True, print the explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic

        Returns:
            list: explanation texts.
        """
        out = np.array(self.operands)
        parent_weights_mask = self._produce_weights_mask(parent_weights.squeeze(), quantile, threshold)

        if parent_mask is not None:
            out = out[parent_mask]
            if isinstance(out, str):
                out = np.array(out)
            if not global_explain:
                input_features = input_features[parent_mask]

        # ensure all are iterable
        out = out.reshape(-1, )
        parent_weights = parent_weights.reshape(-1, )
        parent_weights_mask = parent_weights_mask.reshape(-1, )

        output = []
        for i, (name, weight, pwm) in enumerate(zip(out, parent_weights, parent_weights_mask)):
            if pwm:
                if global_explain:
                    input_features = self._compute_required_inputs(
                        parent_weights=parent_weights,
                        required_output_threshold=required_output_thresholds,
                        parent_logic_type=parent_logic_type,
                        negate=negate,
                        force_negate=force_negate,
                        rounding_precision=rounding_precision
                    )
                    co = input_features[i].squeeze()
                    input_features = input_features.squeeze()
                else:
                    co = input_features[i].squeeze()

                # current_outputs_mask, next_required_output_thresholds = self._produce_current_outputs_mask(
                current_outputs_mask, next_required_output_thresholds, current_outputs_negated, other_outputs = self._produce_current_outputs_mask(
                    co,
                    input_features,
                    parent_weights,
                    parent_weights[i],
                    parent_logic_type,
                    required_output_thresholds,
                    negate,
                    force_negate,
                    rounding_precision
                )

                if feature_importances and feature_importances_type in ['weight', 'weight_proportion']:
                    current_outputs_mask = True

                if feature_importances:
                    importance_value = self._generate_feature_importance(
                        current_outputs_negated=current_outputs_negated,
                        parent_weight=parent_weights[i],
                        feature_importances_type=feature_importances_type,
                        parent_logic_type=parent_logic_type,
                        required_output_thresholds=required_output_thresholds,
                        other_outputs=other_outputs
                    )

                if current_outputs_mask:
                    current_output = self._produce_negation_string(
                        weight, self._produce_predicate_value(
                            name, weight, float(co), float(next_required_output_thresholds), ignore_uninformative,
                            print_type, rounding_precision, negate, inverse_transform, show_bounds, **kwargs),
                        explain_type=explain_type, print_type=print_type, negate=force_negate)
                    if feature_importances:
                        if current_output[-1] == ')':
                            current_output = current_output[:-1] + ' [' + str(importance_value) + '])'
                        else:
                            current_output += ' [' + str(importance_value) + ']'
                    output += [current_output]

        output = np.unique([x for x in output if x.replace("\n", "").replace("\t", "") != "NOT()" and x != ""]).tolist()
        if feature_importances:
            output = [x for x in output if x.replace("\n", "").replace("\t", "") != "NOT()" and x != ""]
        
        if print_explanation:
            explanation_str = ', '.join(output)
            if input_features is not None:
                print(f"Logic at depth {depth}: {explanation_str}"
                      f"\noutput: {input_features.squeeze()}\n")
            else:
                print(f"Logic at depth {depth}: {explanation_str}\n")
        
        if print_type == 'natural':
            if parent_logic_type == 'And':
                output = [', and '.join(output)]
            elif parent_logic_type == 'Or':
                output = [', or '.join(output)]

        return output

    def explain(
            self,
            quantile: float = 0.5,
            required_output_thresholds: torch.Tensor = None,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            negate: bool = False,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            depth: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            feature_importance=False, # TODO check if it should be False by default
            **kwargs
    ) -> list:
        """
        Produce a global explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform

        Returns:
            list: explanation texts.
        """
        return self.explain_sample(
            quantile=quantile,
            required_output_thresholds=required_output_thresholds,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            explain_type=explain_type,
            parent_logic_type=parent_logic_type,
            negate=negate,
            input_features=input_features,
            force_negate=force_negate,
            depth=depth,
            global_explain=True,
            ignore_uninformative=ignore_uninformative,
            print_type=print_type,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            feature_importance=feature_importance,
            **kwargs
        )

    def print_sample(
            self,
            quantile: float = 0.5,
            required_output_thresholds: torch.Tensor = None,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            negate: bool = False,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            depth: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            **kwargs
    ) -> list:
        """
        Print a sample explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
        """
        return self.explain_sample(
            quantile=quantile,
            required_output_thresholds=required_output_thresholds,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            explain_type=explain_type,
            parent_logic_type=parent_logic_type,
            negate=negate,
            input_features=input_features,
            force_negate=force_negate,
            depth=depth,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            print_type=print_type,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            **kwargs
        )

    def print(
            self,
            quantile: float = 0.5,
            required_output_thresholds: torch.Tensor = None,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            negate: bool = False,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            depth: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            **kwargs
    ) -> list:
        """
        Produce a global explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
        """
        return self.explain_sample(
            quantile=quantile,
            required_output_thresholds=required_output_thresholds,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            explain_type=explain_type,
            parent_logic_type=parent_logic_type,
            negate=negate,
            input_features=input_features,
            force_negate=force_negate,
            depth=depth,
            global_explain=True,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            print_type=print_type,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            **kwargs
        )



import logging
from copy import deepcopy

import torch
from torch.utils.data import DataLoader

from nrn.models.base import ReasoningNetworkModel


class BaseReasoningNetworkDistributedTrainer(object):

    def __init__(
            self,
            model: ReasoningNetworkModel,
            loss_func,
            optimizer,
            scheduler=None,
            epochs: int = 100,
            l1_lambda: float = 1e-5,
            early_stopping_plateau_count: int = 20,
            accumulation_steps: int = 1,
            objective: str = 'maximize'
    ):
        """
        Train a ReasoningNetwork model.

        Args:
            model (ReasoningNetwork): the model to be trained.
                Format:
                    This model must implement the following methods:
                        predict(DataLoader): returns Tuple[pd.DataFrame, pd.DataFrame]: prediction values, target values
                        evaluate(DataLoader): returns float: evaluation metric
            loss_func (Loss): One of pytorch's loss classes
            optimizer (Optimizer): One of pytorch's optimizer classes
            scheduler (Scheduler): One of pytorch's scheduler classes
            class_weights (torch.Tensor): tensor of class weights.
            epochs (int): Maximum number of epochs to train for
            l1_lambda (float): L1 regularization weight
            early_stopping_plateau_count (int): Number of epochs without improvement to end training.
            accumulation_steps (int): Number of optimization steps to perform before backward operation.
            validation_steps: Optional.  Number of optimization steps to perform evaluation after.
            objective (str): 'maximize' or 'minimize'.  Default is 'maximize'.
        """
        self.model = model
        self.loss_func = loss_func
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.epochs = epochs
        self.l1_lambda = l1_lambda
        self.early_stopping_plateau_count = early_stopping_plateau_count
        self.accumulation_steps = accumulation_steps
        self.objective = objective

        self.epoch = 0
        if self.objective == 'maximize':
            self.best_val_performance = 0
            self.best_train_performance = 1e12
        elif self.objective == 'minimize':
            self.best_val_performance = 1e12
            self.best_train_performance = 1e12

        self.initialized_optimizer = None
        self.was_pruned = False

        self.USE_CUDA = torch.cuda.is_available()
        self.USE_MPS = torch.backends.mps.is_available()
        if hasattr(self.model, 'rn'):
            self.USE_DATA_PARALLEL = isinstance(self.model.rn, torch.nn.DataParallel)
        else:
            self.USE_DATA_PARALLEL = False

        self.logger = logging.getLogger(self.__class__.__name__)

    def _validate_state_dicts(self, model_state_dict_1, model_state_dict_2):
        if len(model_state_dict_1) != len(model_state_dict_2):
            self.logger.info(
                f"Length mismatch: {len(model_state_dict_1)}, {len(model_state_dict_2)}"
            )
            return False

        # Replicate modules have "module" attached to their keys, so strip these off when comparing to local model.
        if next(iter(model_state_dict_1.keys())).startswith("module"):
            model_state_dict_1 = {
                k[len("module") + 1:]: v for k, v in model_state_dict_1.items()
            }

        if next(iter(model_state_dict_2.keys())).startswith("module"):
            model_state_dict_2 = {
                k[len("module") + 1:]: v for k, v in model_state_dict_2.items()
            }

        for ((k_1, v_1), (k_2, v_2)) in zip(
                model_state_dict_1.items(), model_state_dict_2.items()
        ):
            if k_1 != k_2:
                self.logger.info(f"Key mismatch: {k_1} vs {k_2}")
                return False
            # convert both to the same CUDA device
            if str(v_1.device) != "cuda:0":
                v_1 = v_1.to("cuda:0" if torch.cuda.is_available() else "cpu")
            if str(v_2.device) != "cuda:0":
                v_2 = v_2.to("cuda:0" if torch.cuda.is_available() else "cpu")

            if not torch.allclose(v_1, v_2):
                self.logger.info(f"Tensor mismatch: {v_1} vs {v_2}")
                return False

        self.logger.info("ALL CHECKS ON STATE DICT PASSED")
        return True

    def save_best_state(self):
        """
        Save the state of the network.
        """
        self.model.best_state['rn'] = deepcopy(self.model.rn)
        self.model.best_state['epoch'] = self.epoch
        # self.model.best_state['initialized_optimizer'] = deepcopy(self.optimizer)
        self.model.best_state['was_pruned'] = self.was_pruned

    def set_best_state(self):
        """
        Set the best state from the last saved.
        """
        self.model.rn = deepcopy(self.model.best_state['rn'])
        self.epoch = self.model.best_state['epoch']
        # self.optimizer = deepcopy(self.model.best_state['initialized_optimizer'])
        self.was_pruned = self.model.best_state['was_pruned']

        return self._validate_state_dicts(self.model.rn.state_dict(), self.model.best_state['rn'].state_dict())

    def _rn_to_cuda(self):
        """
        Send model to cuda.
        """
        # to cuda again
        if self.USE_DATA_PARALLEL:
            self.model.rn = torch.nn.DataParallel(self.model.rn.module)
        if self.USE_CUDA:
            self.model.rn = self.model.rn.cuda()
        elif self.USE_MPS:
            self.model.rn = self.model.rn.to('mps')

    def _check_improvement(self, val_performance):
        """
        Check if performances have improved
        """
        if self.objective == 'maximize':
            return val_performance > self.best_val_performance
        elif self.objective == 'minimize':
            return val_performance < self.best_val_performance
        else:
            raise AssertionError("`objective` must be one of 'maximize' or 'minimize'")

    def evaluate_step(self, dl: DataLoader, epoch, plateau_counter, **kwds):
        """
        Perform evaluation and increment counters.

        Args:
            dl (DataLoader): data loader that should be used for evaluation
            epoch (int): current epoch
            plateau_counter (int): plateau_counter for early stopping

        Returns:
            int: plateau_counter
        """
        self.logger.info(f"VALIDATION EVALUATION - EPOCH [{epoch}]: \n")
        predictions, targets = self.model.predict(dl)
        val_performance = self.model.evaluate(
            predictions=predictions,
            labels=targets,
            output_metric=kwds.get('evaluation_metric')
        )
        if self._check_improvement(val_performance):
            self.logger.info(f"CURRENT VALIDATION PERFORMANCE {val_performance} "
                             f"IMPROVED VS BEST VAL PERFORMANCE {self.best_val_performance} "
                             f"UPDATING BEST PARAMETERS.")
            self.best_val_performance = val_performance
            self.save_best_state()
            plateau_counter = 0
        else:
            plateau_counter += 1

        return plateau_counter

    def _validation_step(self, **kwargs):
        plateau_counter = self.evaluate_step(
            dl=kwargs.get('val_dl'), epoch=kwargs.get('epoch'), plateau_counter=kwargs.get('plateau_counter'))
        return {'plateau_counter': plateau_counter}

    def _process_batches(
            self,
            dl: DataLoader,
            optimizer,
            **kwargs,
    ):
        """
        Perform batched epoch.

        Args:
            dl (DataLoader): the training dataloader
                Format:
                    DataLoader must return batches with the keys:
                        'features': [BATCH_SIZE, N_FEATURES]
                        'target': [BATCH_SIZE, N_TARGETS]
            optimizer (Optimizer): pytorch optimizer

        Returns:
            Tuple[float, Tensor]: total loss, final batch of features used for sample level explanation
        """
        total_loss = 0
        total_steps = kwargs.get('total_steps')
        for i, batch in enumerate(dl):

            # [BATCH_SIZE, N_FEATURES]
            features = batch['features']
            # [BATCH_SIZE, N_TARGETS]
            target = batch['target'].squeeze(-1)

            if self.USE_CUDA:
                features = features.cuda()
                target = target.cuda()
            elif self.USE_MPS:
                features = features.to('mps')
                target = target.to('mps')

            # [BATCH_SIZE, N_TARGETS]
            predictions = self.model.rn(features)
            if predictions.ndim > 1:
                predictions = predictions.squeeze(-1)
            # [BATCH_SIZE, N_TARGETS]
            loss = self.loss_func(predictions, target.float())

            if loss.ndim >= 1:
                loss = loss.sum(axis=0)
                loss_agg = loss.mean()
            else:
                loss_agg = loss

            # add l1 loss
            if self.l1_lambda > 0.0:
                l1_norm = sum(p.abs().sum() for p in self.model.rn.parameters())
                loss_agg = loss_agg + self.l1_lambda * l1_norm

            loss_agg /= self.accumulation_steps

            loss_agg.backward()

            if (i + 1) % self.accumulation_steps == 0 or (i + 1 == len(dl)):
                optimizer.step()
                optimizer.zero_grad()
                if self.scheduler is not None:
                    self.scheduler.step()
                total_steps += 1

            # report aggregated loss if using distributed training
            total_loss += loss

        if isinstance(total_loss, torch.Tensor):
            total_loss = total_loss.detach().cpu().numpy() / len(dl.dataset)

        return total_loss, total_steps

    def train(self, train_dl: DataLoader, val_dl: DataLoader = None, evaluation_metric=None, multi_class=False):
        """
        Train the rn model.

        Args:
            train_dl (DataLoader): data loader for training data
            val_dl (DataLoader): data loader for a held-out set used for early stopping
            evaluation_metric (function): scikit-learn evaluation metrics
            multi_class (bool): treat problem as multi-class for evaluation purposes

                Format:
                    DataLoaders must return batches with the keys:
                        'features': [BATCH_SIZE, N_FEATURES]
                        'target': [BATCH_SIZE, N_TARGETS]
        """
        plateau_counter = 0
        total_steps = 0

        if val_dl is None:
            val_dl = train_dl

        for epoch in range(self.epochs):

            self.epoch = epoch

            if epoch > 0:
                plateau_counter = self.evaluate_step(
                    val_dl,
                    epoch,
                    plateau_counter,
                    evaluation_function=evaluation_metric,
                    multi_class=multi_class
                )

            if plateau_counter >= self.early_stopping_plateau_count:
                self.logger.info(f"STOPPING EARLY.  REACHED PLATEAU COUNT OF {plateau_counter}.")
                break

            total_loss, total_steps = self._process_batches(
                dl=train_dl, optimizer=self.optimizer, val_dl=val_dl, epoch=epoch,
                plateau_counter=plateau_counter, total_steps=total_steps)

            self.logger.info(f"Epoch: {epoch}, Total Loss: {total_loss}")
            epoch += 1


__all__ = [BaseReasoningNetworkDistributedTrainer]



from typing import Union

import torch
import torch.distributed as dist


def grad_agg(params):
    r"""Aggregate gradients across multiple processes doing training."""
    for p in params:
        if p.requires_grad:
            world_size = dist.get_world_size()
            val = torch.tensor([float(p.grad is not None)])
            votes = float_agg(val)
            assert (votes == world_size) or (votes == 0)
            if p.grad is not None:
                dist.all_reduce(p.grad.data, op=dist.ReduceOp.SUM)


def float_agg(val):
    r"""Aggregate any value val across multiple processes."""
    val = torch.tensor([val])
    dist.all_reduce(val, op=dist.ReduceOp.SUM)
    return val.item()


def tensor_agg(tensor_in: Union[torch.FloatTensor, torch.LongTensor, torch.DoubleTensor, torch.Tensor]):
    r"""Aggregate any tensor by concatenation across multiple processes"""
    world_size = dist.get_world_size()
    tensor_out = [torch.zeros_like(tensor_in.reshape(-1)) for _ in range(world_size)]
    dist.all_gather(tensor_out, tensor_in.reshape(-1))
    tensor_out = torch.cat(tensor_out)
    return tensor_out


__all__ = [grad_agg, float_agg, tensor_agg]


import logging
from typing import List

import torch

from ..models.base import BoostedBanditNRNModel
from .mixins import ReasoningNetworkClassifierMixin


class BanditNRNClassifier(BoostedBanditNRNModel, ReasoningNetworkClassifierMixin):

    def __init__(
            self,
            target_names: List[str],
            feature_names: List[str],
            input_size: int,
            output_size: int,
            layer_sizes: List[int],
            n_selected_features_input: int,
            n_selected_features_internal: int,
            n_selected_features_output: int,
            perform_prune_quantile: float,
            ucb_scale: float,
            normal_form: str = 'dnf',
            delta: float = 2.0,
            prune_strategy: str = 'class',
            bootstrap: bool = True,
            swa: bool = True,
            add_negations: bool = False,
            weight_init: float = 0.2,
            policy_init: torch.Tensor = None,
            xgb_max_depth: int = 4,
            xgb_n_estimators: int = 200,
            xgb_min_child_weight: float = 1.0,
            xgb_subsample: float = 0.7,
            xgb_learning_rate: float = 0.001,
            xgb_colsample_bylevel: float = 0.7,
            xgb_colsample_bytree: float = 0.7,
            xgb_gamma: float = 1,
            xgb_reg_lambda: float = 2.0,
            xgb_reg_alpha: float = 2.0,
            logits: bool = False
    ):
        """
        Initialize a BERrnClassifier model.

        Example:
            model = BERrnClassifier(
                target_names=['class1', 'class2'],
                feature_names=['feature1', 'feature2', 'feature3'],
                input_size=3,
                output_size=2,
                layer_sizes=[3, 3]
                n_selected_features_input=2,
                n_selected_features_internal=2,
                n_selected_features_output=1,
                ucb_scale=1.96,
                perform_prune_quantile=0.7,
                normal_form='dnf',
                prune_strategy='class',
                bootstrap=True,
                distributed=False
            )

        Args:
            target_names (list): A list of the target names.
            feature_names (list): A list of feature names.
            input_size (int): number of features from input.
            output_size (int): number of outputs.
            layer_sizes (list): A list containing the number of output logics for each layer.
            n_selected_features_input (int): The number of features to include in each logic in the input layer.
            n_selected_features_internal (int): The number of logics to include in each logic in the internal layers.
            n_selected_features_output (int): The number of logics to include in each logic in the output layer.
            perform_prune_quantile (float): The quantile to use for pruning randomized rn.
            ucb_scale (float): The scale of the confidence interval in the multi-armed bandit policy.
                               c = 1.96 is a 95% confidence interval.
            normal_form (str): 'dnf' for disjunctive normal form network; 'cnf' for conjunctive normal form network.
            delta (float): higher values increase diversity of logic generation away from existing logics.
            prune_strategy(str): Either 'class' or 'logic'.  Determines which pruning strategy to use.
            bootstrap (bool): Use boostrap samples to evaluate each atomic logic in logic prune strategy.
            swa (bool): Use stochastic weight averaging
            add_negations (bool): add negations of logic.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
            xgb_max_depth (int): Max depth for XGBoost boosting model.
            xgb_n_estimators (int): Number of estimators for XGBoost boosting model.
            xgb_min_child_weight (float): Minimum child weight for XGBoost boosting model.
            xgb_subsample (float): Subsample percentage for XGBoost boosting model.
            xgb_learning_rate (float): Learning rate for XGBoost boosting model.
            xgb_colsample_bylevel (float): Column subsample percent for XGBoost boosting model.
            xgb_colsample_bytree (float): Tree subsample percent for XGBoost boosting model.
            xgb_gamma (float): Gamma parameter for XGBoost boosting model.
            xgb_reg_lambda (float): Lambda regularization parameter for XGBoost boosting model.
            xgb_reg_alpha (float): Alpha regularization parameter for XGBoost boosting model.
        """
        ReasoningNetworkClassifierMixin.__init__(self, output_size=output_size)
        BoostedBanditNRNModel.__init__(
            self,
            target_names=target_names,
            input_size=input_size,
            output_size=output_size,
            layer_sizes=layer_sizes,
            feature_names=feature_names,
            n_selected_features_input=n_selected_features_input,
            n_selected_features_internal=n_selected_features_internal,
            n_selected_features_output=n_selected_features_output,
            perform_prune_quantile=perform_prune_quantile,
            ucb_scale=ucb_scale,
            normal_form=normal_form,
            delta=delta,
            prune_strategy=prune_strategy,
            bootstrap=bootstrap,
            swa=swa,
            weight_init=weight_init,
            policy_init=policy_init,
            add_negations=add_negations,
            xgb_max_depth=xgb_max_depth,
            xgb_n_estimators=xgb_n_estimators,
            xgb_min_child_weight=xgb_min_child_weight,
            xgb_subsample=xgb_subsample,
            xgb_learning_rate=xgb_learning_rate,
            xgb_colsample_bylevel=xgb_colsample_bylevel,
            xgb_colsample_bytree=xgb_colsample_bytree,
            xgb_gamma=xgb_gamma,
            xgb_reg_lambda=xgb_reg_lambda,
            xgb_reg_alpha=xgb_reg_alpha,
            logits=logits
        )
        self.set_modules(self.rn)
        self.logger = logging.getLogger(self.__class__.__name__)


__all__ = [BanditNRNClassifier]



import logging
from copy import deepcopy
from typing import Union

import numpy as np

import torch
from torch.nn import Parameter, functional

from .predicates import BasePredicates
from ._core import LukasiewiczCore
from nrn.utils.operations import val_clamp, EXU


class LukasiewiczChannelBlock(LukasiewiczCore):

    EPS = 1e-3

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands,
            logic_type: str,
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2
    ):
        """
        A logical and channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            logic_type (str): type of logic for self.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
        """
        super(LukasiewiczChannelBlock, self).__init__()
        size = (channels, n_selected_features, out_features)
        weights = torch.ones(size).uniform_(-1 * weight_init, weight_init)
        if add_negations:
            weights = torch.cat([weights, weights.clone() * -1], dim=1)

        self.weights = Parameter(weights)
        self.bias = Parameter(torch.tensor(1.0), requires_grad=False)

        self.channels = channels
        self.in_features = in_features
        self.out_features = out_features
        self.operands = operands
        self.outputs_key = outputs_key
        self.parent_weights_dimension = parent_weights_dimension
        self.n_selected_features = n_selected_features
        self.logic_type = logic_type
        self.knowledge_added = False
        self.add_negations = add_negations
        self.weight_init = weight_init

        # MASK IS SIZE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES] and values range from 0 to IN_FEATURES - 1
        mask = torch.empty(
            (channels, out_features, in_features)).uniform_(0, 1).topk(n_selected_features, dim=-1).indices
        if add_negations:
            mask = torch.cat([mask, mask.clone()], dim=-1)
        self.register_buffer('mask', mask)
        self.logger = logging.getLogger(self.__class__.__name__)

    def explain_sample(
            self,
            required_output_thresholds: torch.Tensor,
            outputs_dict: dict = None,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            channel: int = 0,
            global_explain: bool = False,
            print_explanation: bool = False,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            feature_importances: bool = False,
            feature_importances_type: str = 'weight',
            **kwargs
    ) -> list:
        """
        Produce a sample explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            outputs_dict (dict): Dictionary of outputs from forward pass.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            global_explain (bool): If True, perform a global explanation.
            print_explanation (bool): If True, print the explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic

        Returns:
            list: explanation texts.
        """
        if isinstance(parent_weights, float) or parent_weights.ndim == 0:
            parent_weights = torch.tensor([parent_weights])

        if not global_explain:
            current_outputs = outputs_dict[self.outputs_key]
            od = deepcopy(outputs_dict)
        else:
            od = None

        if hasattr(self, 'var_emb_dim') and hasattr(self, 'var_n_layers'):
            # variational
            mask = self.sample_mask()
            weights_to_use, mask_to_use = self.weights * mask, self.mask
        elif hasattr(self, 'attn_emb_dim') and hasattr(self, 'attn_n_layers'):
            if not isinstance(self.operands, BasePredicates):
                current_inputs = outputs_dict[self.operands.outputs_key]
            else:
                current_inputs = input_features.to(self.weights.device)
                if current_inputs.ndim == 1:
                    current_inputs = current_inputs.unsqueeze(0)
            weights_to_use, mask_to_use = self.produce_explanation_weights(current_inputs), self.mask
        else:
            weights_to_use, mask_to_use = self.weights, self.mask

        # select ranges that need to be processed
        out_feature_range = np.arange(self.out_features)
        parent_weights_mask = self._produce_weights_mask(parent_weights, quantile, threshold)

        if parent_mask is not None:
            out_feature_range = out_feature_range[parent_mask]
        if isinstance(out_feature_range, np.int64):
            out_feature_range = torch.tensor([out_feature_range])

        # process ranges
        explanation = []
        for i, (out_feature, parent_weight_mask) in enumerate(zip(out_feature_range, parent_weights_mask)):
            if parent_weight_mask:  # if the logic is above the quantile threshold then produce the explanation
                if global_explain:
                    oo = self._compute_required_inputs(
                        parent_weights=parent_weights,
                        required_output_threshold=min(
                            required_output_thresholds * (1 - self.EPS if force_negate else 1 + self.EPS), 1.0),
                        # required_output_threshold=required_output_thresholds,
                        parent_logic_type=parent_logic_type,
                        negate=negate,
                        rounding_precision=rounding_precision
                    )
                    co = oo[i].squeeze()
                    oo = oo.squeeze()
                else:
                    co = current_outputs[0, channel, 0, out_feature].cpu()
                    oo = current_outputs[0, channel, 0, parent_mask].cpu().squeeze()

                # current_outputs_mask, next_required_output_thresholds = self._produce_current_outputs_mask(
                current_outputs_mask, next_required_output_thresholds, current_outputs_negated, other_outputs = self._produce_current_outputs_mask(
                    co,
                    oo,
                    parent_weights,
                    parent_weights[i],
                    parent_logic_type,
                    required_output_thresholds,
                    negate,
                    force_negate,
                    rounding_precision
                )

                if feature_importances and feature_importances_type in ['weight', 'weight_proportion']:
                    current_outputs_mask = True

                # if current_outputs_mask or print_explanation:
                if current_outputs_mask:

                    negation = negate if parent_weights[i] >= 0. else not negate
                    op_explain = self.operands.explain_sample(
                        outputs_dict=od,
                        required_output_thresholds=next_required_output_thresholds,
                        quantile=quantile,
                        threshold=threshold,
                        parent_weights=weights_to_use[channel, :, out_feature].detach().cpu().squeeze(),
                        parent_mask=mask_to_use[channel, out_feature, :].detach().cpu(),
                        parent_logic_type=self.logic_type,
                        negate=negation,
                        depth=depth + 1,
                        explain_type=explain_type,
                        print_type=print_type,
                        input_features=input_features,
                        channel=channel,
                        global_explain=global_explain,
                        print_explanation=print_explanation,
                        ignore_uninformative=ignore_uninformative,
                        rounding_precision=rounding_precision,
                        inverse_transform=inverse_transform,
                        force_negate=force_negate,
                        show_bounds=show_bounds,
                        feature_importances=feature_importances,
                        feature_importances_type=feature_importances_type,
                        **kwargs
                    )

                    if op_explain:
                        # TODO find a better name for tmp
                        tmp = self._produce_negation_string(
                                parent_weights[i],
                                f'{self._produce_logic_string(op_explain, print_type, depth)}',
                                explain_type=explain_type,
                                print_type=print_type,
                                depth=depth,
                            )

                        if feature_importances:
                            importance_value = self._generate_feature_importance(
                                    current_outputs_negated=current_outputs_negated,
                                    parent_weight=parent_weights[i],
                                    feature_importances_type=feature_importances_type,
                                    parent_logic_type=parent_logic_type,
                                    required_output_thresholds=required_output_thresholds,
                                    other_outputs=other_outputs
                            )

                            explanation_str = \
                                self._produce_negation_string(
                                    parent_weights[i],
                                    f'{self._produce_logic_string(op_explain, print_type, depth)}',
                                    explain_type=explain_type,
                                    print_type=print_type,
                                    depth=depth,
                                )
                            # TODO could there be multiple consecutive NOT we may need to remove?
                            if explanation_str.startswith('NOT(NOT(') and (explanation_str.endswith('))')):
                                explanation_str = explanation_str[8:-2]
                                # print("removed NOT NOT:", explanation_str)
                            if (explanation_str.startswith('NOT(AND')) and (explanation_str.endswith('))')):
                                # print("NOT AND")
                                if feature_importances_type in ['weight_proportion']:
                                    tmp = 'NOT(AND[' + str(importance_value) + ']' + explanation_str[7:]
                                    # tmp = 'NOT(AND[1.0]' + explanation_str[7:]
                                else:
                                    tmp = 'NOT(AND[' + str(importance_value) + ']' + explanation_str[7:]
                            elif (explanation_str.startswith('NOT(OR')) and (explanation_str.endswith('))')):
                                # print("NOT OR")
                                if feature_importances_type in ['weight_proportion']:
                                    tmp = 'NOT(OR[' + str(importance_value) + ']' + explanation_str[6:]
                                    # tmp = 'NOT(AND[1.0]' + explanation_str[7:]
                                else:
                                    tmp = 'NOT(OR[' + str(importance_value) + ']' + explanation_str[6:]
                            elif (explanation_str.startswith('AND')):
                                tmp = 'AND[' + str(importance_value) + ']' + explanation_str[3:]
                            elif (explanation_str.startswith('OR')):
                                tmp = 'OR[' + str(importance_value) + ']' + explanation_str[2:]
                            explanation += [tmp]
                        else:
                            explanation_str = [
                                self._produce_negation_string(
                                    parent_weights[i],
                                    f'{self._produce_logic_string(op_explain, print_type, depth)}',
                                    explain_type=explain_type,
                                    print_type=print_type,
                                    depth=depth
                                )]
                            explanation += explanation_str

                        if print_explanation:
                            print(f"Logic at depth {depth}: {explanation_str}"
                                  f"\nweights: {weights_to_use[channel, :, out_feature].detach().cpu().squeeze()}"
                                  f"\noutput: {co if parent_weights[i] >= 0 else 1. - co}"
                                  f"\nrequired_threshold: {next_required_output_thresholds}\n")

        explanation = np.unique([x for x in explanation if
                                 x != "" and x.replace("\t", "").replace("\n", "") != 'NOT()']).tolist()
        if feature_importances:
            explanation = [x for x in explanation if
                                 x != "" and x.replace("\t", "").replace("\n", "") != 'NOT()']

        if explanation:
            return explanation
        return [""]

    def explain(
            self,
            required_output_thresholds: torch.Tensor,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Produce a global explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            show_bounds (bool): include numeric boundary used in logic

        Returns:
            list: explanation texts.
        """
        return self.explain_sample(
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            force_negate=force_negate,
            channel=channel,
            global_explain=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )

    def print_sample(
            self,
            outputs_dict: dict,
            required_output_thresholds: torch.Tensor,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Print a sample explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            outputs_dict (dict): Dictionary of outputs from forward pass.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            show_bounds (bool): include numeric boundary used in logic
        """
        return self.explain_sample(
            outputs_dict=outputs_dict,
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            input_features=input_features,
            force_negate=force_negate,
            channel=channel,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )

    def print(
            self,
            required_output_thresholds: torch.Tensor,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: torch.Tensor = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Produce a global explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            show_bounds (bool): include numeric boundary used in logic
        """
        return self.explain_sample(
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            force_negate=force_negate,
            channel=channel,
            global_explain=True,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )

    def add_knowledge(
            self,
            channel,
            out_feature,
            input_indices,
            required_thresholds,
            required_negations,
            freeze_knowledge=True
    ):
        # if knowledge has not been added yet then zero the weights
        if not self.knowledge_added:
            self.weights.data.copy_(torch.zeros_like(self.weights.data))

        # update the mask to use the given predicates
        # [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        self.mask.data[channel, out_feature, :] = torch.tensor(input_indices)

        # update the weights to use the given required thresholds
        # [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]

        # flip negated thresholds
        required_thresholds = torch.tensor(required_thresholds)
        required_negations = torch.tensor(required_negations)
        required_thresholds = (torch.relu(required_negations) * required_thresholds
                               + (1 - torch.relu(required_negations)) * (1 - required_thresholds))

        # determine weights
        non_zero_weight_count = (required_thresholds >= 0).sum()
        if self.logic_type == 'Or':
            required_weights = (1.0/non_zero_weight_count) / required_thresholds
        elif self.logic_type == 'And':
            required_weights = (1.0/non_zero_weight_count) / (1.0 - required_thresholds)
        else:
            raise AssertionError("logic_type must be 'And', 'Or'")
        required_weights = required_weights * required_negations
        required_weights = required_weights * (required_thresholds > 0.0)
        required_weights[torch.logical_and(required_weights >= -1e-8, required_weights <= 1e-8)] = 0.0

        self.weights.data[channel, :, out_feature] = required_weights
        self.weights.requires_grad_(not freeze_knowledge)

        self.knowledge_added = True


class VariationalLukasiewiczChannelBlock(LukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands,
            logic_type: str,
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            var_emb_dim: int = 50,
            var_n_layers: int = 2
    ):
        """
        A logical and channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            logic_type (str): type of logic for self.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
        """
        super(VariationalLukasiewiczChannelBlock, self).__init__(
            channels=channels,
            in_features=in_features,
            out_features=out_features,
            n_selected_features=n_selected_features,
            parent_weights_dimension=parent_weights_dimension,
            operands=operands,
            logic_type=logic_type,
            outputs_key=outputs_key,
            add_negations=add_negations,
            weight_init=weight_init
        )
        self.var_emb_dim = var_emb_dim
        self.var_n_layers = var_n_layers
        assert self.var_n_layers >= 2, "`attn_n_layers` must be at least 1."

        self.var_emb = torch.randn(size=(channels * out_features, var_emb_dim))

        # construct variational mean network
        var_mean = [torch.nn.Linear(self.var_emb_dim, self.var_emb_dim), torch.nn.ReLU()]
        for _ in range(self.var_n_layers - 2):
            var_mean += [torch.nn.Linear(self.var_emb_dim, self.var_emb_dim), torch.nn.ReLU()]
        var_mean += [torch.nn.Linear(self.var_emb_dim, self.weights.size(1))]
        self.var_mean = torch.nn.Sequential(*var_mean)

        # construct variational log std network
        var_std = [torch.nn.Linear(self.var_emb_dim, self.var_emb_dim), torch.nn.ReLU()]
        for _ in range(self.var_n_layers - 2):
            var_std += [torch.nn.Linear(self.var_emb_dim, self.var_emb_dim), torch.nn.ReLU()]
        var_std += [torch.nn.Linear(self.var_emb_dim, self.weights.size(1))]
        self.var_std = torch.nn.Sequential(*var_std)

        self.N = torch.distributions.Normal(0, 1)
        self.N.loc = self.N.loc
        self.N.scale = self.N.scale

        self.tau = Parameter(torch.ones(size=(self.weights.size(0), 1, self.weights.size(2)))).float()

        self.seed_count = 0

    def sample_mask(self):
        # variational sparsity
        if not self.training:
            torch.manual_seed(0)
            torch.cuda.manual_seed(0)
            torch.cuda.manual_seed_all(0)  # if you are using multi-GPU.
        else:
            self.seed_count += 1
            torch.manual_seed(self.seed_count)
            torch.cuda.manual_seed(self.seed_count)
            torch.cuda.manual_seed_all(self.seed_count)  # if you are using multi-GPU.

        # If the distributions diverge then reset to draw from a normal dist with mean 1. std 1.
        mu = torch.nan_to_num(self.var_mean(self.var_emb.to(self.var_mean[0].weight.device)), 0., 0., 0.)
        sigma = torch.nan_to_num(
            torch.exp(self.var_std(self.var_emb.to(self.var_mean[0].weight.device))), 1., 1., 1.)
        z = mu + sigma * self.N.sample(mu.shape).to(self.var_mean[0].weight.device)
        z = z.reshape(self.weights.size(0), self.weights.size(2), self.weights.size(1)).transpose(2, 1)
        mask = functional.gumbel_softmax(
            torch.logit(z, eps=1e-6),  # how else can we get tau not to diverge?
            dim=1,
            tau=val_clamp(self.tau, _min=0.1, _max=1.0).float()
        )
        mask = torch.nan_to_num(mask, 0., 0., 0.)  # how do we avoid divergence?
        return mask


class AttentionLukasiewiczChannelBlock(LukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands,
            logic_type: str,
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = None,
            attn_emb_dim: int = 32,
            attn_n_layers: int = 2
    ):
        """
        A logical and channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            logic_type (str): type of logic for self.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
        """
        super(AttentionLukasiewiczChannelBlock, self).__init__(
            channels=channels,
            in_features=in_features,
            out_features=out_features,
            n_selected_features=n_selected_features,
            parent_weights_dimension=parent_weights_dimension,
            operands=operands,
            logic_type=logic_type,
            outputs_key=outputs_key,
            add_negations=add_negations,
            weight_init=weight_init
        )

        self.attn_emb_dim = attn_emb_dim
        self.attn_n_layers = attn_n_layers
        assert self.attn_n_layers >= 2, "`attn_n_layers` must be at least 1."

        # construct attention network (original)
        attn = [torch.nn.Linear(self.weights.size(1) * 2, self.attn_emb_dim), torch.nn.ReLU()]
        for _ in range(self.attn_n_layers - 2):
            attn += [torch.nn.Linear(self.attn_emb_dim, self.attn_emb_dim), torch.nn.ReLU()]
        attn += [torch.nn.Linear(self.attn_emb_dim, self.weights.size(1))]
        self.attn = torch.nn.Sequential(*attn)
        # end construct attention network (original)

        # # construct attention network (nam like)
        # attn = []
        # for i in range(self.attn_n_layers - 2):
        #     if i == self.attn_n_layers - 3:
        #         attn += [torch.nn.Linear(self.attn_emb_dim, self.attn_emb_dim // 2), torch.nn.ReLU()]
        #     else:
        #         attn += [torch.nn.Linear(self.attn_emb_dim, self.attn_emb_dim), torch.nn.ReLU()]
        # if attn:
        #     attn += [torch.nn.Linear(self.attn_emb_dim // 2, 1, bias=False)]
        # else:
        #     attn = [torch.nn.Linear(self.attn_emb_dim, 1, bias=False)]
        #
        # for layer in attn:
        #     if isinstance(layer, torch.nn.Linear):
        #         torch.nn.init.xavier_uniform_(layer.weight, gain=torch.nn.init.calculate_gain('relu'))
        #
        # self.attn = torch.nn.Sequential(EXU(1, self.attn_emb_dim), *attn)
        # # end construct attention network (nam like)

        self.tau = Parameter(torch.ones(size=(1, self.weights.size(0), self.weights.size(2), 1, 1))).float()

        self.seed_count = 0

    def sample_mask(self, x):

        # variational sparsity
        if not self.training:
            torch.manual_seed(0)
            torch.cuda.manual_seed(0)
            torch.cuda.manual_seed_all(0)  # if you are using multi-GPU.
        else:
            self.seed_count += 1
            torch.manual_seed(self.seed_count)
            torch.cuda.manual_seed(self.seed_count)
            torch.cuda.manual_seed_all(self.seed_count)  # if you are using multi-GPU.

        # apply attention layers (Original)
        h = self.attn(x)
        # end apply attention layers (Original)

        # # apply attention layers NAM like
        # # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
        # x = x.unsqueeze(-1).unsqueeze(-1)
        # # residual connections
        # for i, layer in enumerate(self.attn):
        #     if i == 0:
        #         h = val_clamp(layer(x))
        #     elif i < len(self.attn) - 1:
        #         h = layer(h)
        #     else:
        #         h = layer(h)
        # h = h.squeeze(-1).squeeze(-1)
        # # end apply attention layers NAM like

        attn = functional.gumbel_softmax(
            torch.logit(h, eps=1e-6),  # how else can we get tau not to diverge?
            dim=-1,
            tau=val_clamp(self.tau, _min=0.1, _max=1.0).float()
        )
        attn = torch.nan_to_num(attn, 0., 0., 0.)  # how do we avoid divergence?

        return attn

    def produce_explanation_weights(self, x):
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        x = x.gather(
            -1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        weights_sign_attn = self.weights.data.detach().clone().repeat(
            x.size(0), 1, 1, 1).transpose(-2, -1).unsqueeze(-2).greater_equal(torch.tensor(0.0)).float()
        # pos_attn = x * weights_sign_attn
        # neg_attn = (1 - x) * (1 - weights_sign_attn)
        # x_attn = pos_attn + neg_attn

        mask = self.sample_mask(torch.cat([x, weights_sign_attn], dim=-1))
        mask = mask.squeeze(-2).transpose(-2, -1)
        weights = self.weights.repeat(x.size(0), 1, 1, 1) * mask
        weights = weights.squeeze(0)
        return weights



import torch
from .base import BaseConcatenateBlocksLogic


class ConcatenateBlocksLogic(BaseConcatenateBlocksLogic):

    def __init__(self, modules, outputs_key):
        super(ConcatenateBlocksLogic, self).__init__(modules, outputs_key)

    def forward(self, *inputs):
        """
        A channel logical xor.

        Args:
            *inputs: comma separated tensors of [BATCH_SIZE, CHANNELS, 1, 1]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, # of input tensors]
        """
        return torch.cat(inputs, dim=-1)
